#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# opsidirectoryconnector complements the desktop management solution opsi
# (open pc server integration) http://www.opsi.org
# Copyright (C) 2016-2017 uib GmbH <info@uib.de>

"""
opsi directory connector

Connects LDAP-compatible directories and opsi.

Please visit http://uib.de/ for more information.
"""

import argparse
import logging
import json
import time
from contextlib import contextmanager
from functools import partial

import requests

from ldap3 import Connection
try:
    from ldap3 import LDAPKeyError
except ImportError:
    # Probably ldap3 > 2.0
    from ldap3.core.exceptions import LDAPKeyError

__version__ = '14'
USERAGENT = "opsi directory connector v{0}".format(__version__)
DEFAULT_ATTRIBUTES = set(['dn', 'cn', 'objectCategory', 'objectClass'])
NAME_ADDITION = '_'

logging.basicConfig()
LOGGER = logging.getLogger('opsidirectoryconnector')


class ModulesError(Exception):
    pass


def runConnector():
    cliArguments = parseCommandlineArguments()
    LOGGER.setLevel(cliArguments.logLevel * 10)
    config = readConfig(cliArguments.config)
    config = convertConfig(config)
    rootGroup = config['behaviour']['root_dir_in_opsi']
    LOGGER.debug("root group: {}".format(rootGroup))

    if not rootGroup:
        raise ValueError("Invalid configuration for behaviour -> root_dir_in_opsi")

    if cliArguments.checkOpsi:
        checkConnectionToOpsi(config['opsi'])
        return  # No further processing

    searchQuery = createSearchQuery(
        computerQuery=config['directory'].pop('searchQueryForComputers', None),
        groupQuery=config['directory'].pop('searchQueryForGroups', None),
    )
    elements = readDirectory(mapping=config['mapping'], searchQuery=searchQuery, **config['directory'])
    elements = (convertElementToDict(element) for element in elements)
    if cliArguments.checkDirectory:
        checkConnectionToDirectory(elements)
        return  # No further processing

    writeChanges = config['behaviour'].get('write_changes_to_opsi', False)

    with getSession(**config['opsi']) as opsiSession:
        modulesCheck(opsiSession)
        clientIDs = getIdsOfClientsInOpsi(opsiSession)
        opsiClientGroupMapping = getClientGroupMapping(opsiSession)

        actions = convertElements(
            elements, clientIDs, opsiClientGroupMapping,
            opsiSession, config['mapping'], rootGroup, **config['behaviour']
        )
        updateOpsi(opsiSession, actions, writeChanges=writeChanges)

        opsiSession.get(params=json.dumps({"method": "backend_exit", "params": [], "id": 1}))


def parseCommandlineArguments():
    parser = argparse.ArgumentParser(description="opsi directory connector")
    parser.add_argument('--version', action="version", version=__version__)
    parser.add_argument('--log-level', '-l', dest='logLevel',
                        help=(
                            'Sets how much information will be logged. '
                            'From 1 (debug) to 5 (critical)'
                        ),
                        type=int, choices=[1, 2, 3, 4, 5], default=3)
    checks = parser.add_mutually_exclusive_group()
    checks.add_argument('--check-directory', dest="checkDirectory",
                        action="store_true",
                        help=(
                            'Check if a connection to the directory '
                            'can be established and if items will be '
                            'received.')
                        )
    checks.add_argument('--check-opsi', dest="checkOpsi",
                        action="store_true",
                        help=(
                            'Check if a connection to the opsi server '
                            'can be established.')
                        )
    parser.add_argument('--config', help='Path to the config.', required=True)
    return parser.parse_args()


def readConfig(path):
    with open(path, 'r') as configFile:
        return json.load(configFile)


def convertConfig(config):
    """
    Convert the given `config` into a format that allows passing the \
config to our processing functions.
    """
    mappings = {
        "directory": {
            "search_base": "searchBase",
            "search_query_computers": "searchQueryForComputers",
            "search_query_groups": "searchQueryForGroups",
            "connection_options": "connectionOptions",
        },
        "behaviour": {
            "update_existing_clients": "updateExistingClients",
            "prefer_location_from_directory": "preferLocationFromDrectory",
        },
        "opsi": {
            "connection_options": "connectionOptions",
        }
    }

    newConfig = {}

    for branch, configs in config.items():
        newConfig[branch] = {}

        for key, value in configs.items():
            try:
                newKey = mappings[branch][key]
                newConfig[branch][newKey] = value
            except KeyError:
                newConfig[branch][key] = value

    removePrefixFromMapping(newConfig, "client", "c")
    removePrefixFromMapping(newConfig, "group", "g")

    return newConfig


def modulesCheck(session):
    hostCount = len(getFromOpsi(session, "host_getIdents", ['unicode', {"type": "OpsiClient"}]))
    LOGGER.debug("Clients in opsi: {0}".format(hostCount))
    backendInfo = getFromOpsi(session, "backend_info")
    try:
        modules = backendInfo['modules']
        realmodules = backendInfo.get('realmodules', {})
        if not modules:
            raise ModulesError("No Modules detected.")
        if not modules.get('valid'):
            raise ModulesError('Modules file invalid.')
        elif not modules.get('customer'):
            raise ModulesError('No customer detected in modules file.')
        elif (modules.get('expires', '') != 'never') and (time.mktime(time.strptime(modules.get('expires', '2000-01-01'), "%Y-%m-%d")) - time.time() <= 0):
            raise ModulesError('Modules file expired.')
        elif not modules.get('directory-connector'):
            raise ModulesError('Module directory-connector not available in modules.')

        if realmodules and hostCount:
            countInModules = int(realmodules.get('directory-connector', 0))
            if countInModules and hostCount > countInModules + 50:
                raise ModulesError('Client limit {0} reached. Upgrade your license!'.format(countInModules))
    except Exception as verificationError:
        LOGGER.critical("Missing or invalid licensing detected ({0!s}). Please contact sales@uib.de!".format(verificationError))
        raise verificationError


def removePrefixFromMapping(config, objectType, prefix):
    "Removes the prefix `prefix` from the mapping configued in `config` for `objectType`."
    mapping = config['mapping'][objectType]
    for key in list(mapping.keys()):
        if key.startswith(prefix):
            modifiedKey = key[1:]
            mapping[modifiedKey] = mapping[key]
            del mapping[key]

    config['mapping'][objectType] = mapping


def checkConnectionToOpsi(connectionConfiguration):
    with getSession(**connectionConfiguration) as opsiSession:
        modulesCheck(opsiSession)
        backendInfo = getFromOpsi(opsiSession, 'backend_info')
        LOGGER.info("Connected to opsi service v{0}".format(backendInfo['opsiVersion']))


def readDirectory(address, user, password, searchBase, mapping,
                  searchQuery, connectionOptions={}):

    LOGGER.debug("LDAP search query: {!r}".format(searchQuery))
    attr = DEFAULT_ATTRIBUTES.union(getAdAttributes(mapping))
    LOGGER.debug("Querying for the following attributes: {0}".format(', '.join(attr)))

    with directoryConnection(address, user=user, password=password, connectionOptions=connectionOptions) as conn:
        conn.search(searchBase, searchQuery, attributes=list(attr))
        LOGGER.info('Connected to directory')

        LOGGER.debug('Searching for elements in directory...')
        try:
            for entry in conn.entries:
                LOGGER.debug('Found element: {!r}'.format(entry))
                yield entry
        except AttributeError as attrError:
            LOGGER.debug("Caught attribute error: {0}".format(attrError))
            LOGGER.debug('result: {0}'.format(conn.result))

            LOGGER.debug("Trying fallback for old versions of ldap3...")
            try:
                for entry in conn.response:
                    LOGGER.debug('Found element: {!r}'.format(entry))
                    yield entry
            except AttributeError as fallBackAttrError:
                LOGGER.debug("Fallback failed: {0}".format(fallBackAttrError))

                raise attrError  # We want the original error

        LOGGER.debug("Finished returning elements from directory")


def createSearchQuery(computerQuery=None, groupQuery=None):
    """
    Create a search query for our search.

    :raises ValueError: If no queries are given.
    :returntype: str
    """
    groupQuery = None  # resetting this until the feature is implemented.
    if computerQuery and groupQuery:
        return '(|{}{})'.format(computerQuery,
                                groupQuery)
    elif computerQuery:
        return computerQuery
    elif groupQuery:
        return groupQuery
    else:
        raise ValueError("Missing search configuration!")


def convertElementToDict(element):
    """
    Converts the given `element` to a dict.

    This aims at working with different versions of LDAP3.

    :type element: dict or an element returned from an LDAP search.
    :returntype: dict
    """
    if isinstance(element, dict):
        return element

    try:
        elemDict = json.loads(element.entry_to_json())
    except AttributeError as attErr:
        LOGGER.debug("Unable to get {0!r} as dict: {1!r}".format(element, attErr))
        LOGGER.debug("The element offers the following: {0}".format(dir(element)))

        try:
            elemDict = element.entry_attributes_as_dict()
        except AttributeError as attrErr:
            LOGGER.debug("Failed to call entry_attributes_as_dict: {!r}".format(attrErr))

            elemDict = element.entry_get_attributes_dict()

    LOGGER.debug("elemDict is: {0!r}".format(elemDict))

    if 'dn' not in elemDict:
        LOGGER.debug("Missing DN. Trying to read...")
        try:
            elemDict['dn'] = element.entry_get_dn()
        except AttributeError as attrErr:
            LOGGER.critical("Unable to get DN for element!")
            raise attrErr

    return elemDict


def checkConnectionToDirectory(elements):
    index = 0
    try:
        for index, element in enumerate(elements, start=1):
            LOGGER.debug("Picked up element #{0}: {1}".format(index, element))
        LOGGER.info("The current settings matched {0} elements.".format(index))
    except Exception as connectionError:
        LOGGER.critical("Connecting to the directory failed: {0}".format(connectionError))

    if not index:
        LOGGER.warning("No elements found. Does your search criteria fit?")


def getSession(address, username, password, connectionOptions):
    session = requests.Session()
    session.auth = (username, password)
    session.headers.update({"user-agent": str(USERAGENT)})

    opsiEndpoint = getOpsiEndpointAddress(address)

    # We prepare the get method so that the settings for
    # certificate verification and the appropriate address
    # are always given to avoid passing this to any call.
    verifyCertificate = connectionOptions.get('verify_certificate', True)
    LOGGER.debug("Verify certificate? {}".format(verifyCertificate))
    session.get = partial(session.get, opsiEndpoint, verify=verifyCertificate)

    return session


def getOpsiEndpointAddress(address):
    address = address.strip()
    if not address.endswith('/rpc'):
        if not address.endswith('/'):
            address = address + '/'

        address = address + 'rpc'

    return address


def getAdAttributes(mapping):
    attr = set()
    for objectType, attributeMapping in mapping.items():
        for opsiAttribute, directoryAttribute in attributeMapping.items():
            if directoryAttribute:
                attr.add(directoryAttribute)
            else:
                LOGGER.warning("No value given for opsi {} attribute {!r}".format(objectType, opsiAttribute))

    return attr


@contextmanager
def directoryConnection(address, user, password, connectionOptions):
    if '://' not in address:
        LOGGER.warning("Directory address misses protocol, adding 'ldap://'.")
        address = 'ldap://' + address

    with Connection(address, user=user, password=password) as connection:
        if connectionOptions.get('start_tls', True):
            connection.start_tls()

        if not connection.bind():
            raise RuntimeError("Connection failed: {0}".format(connection.result))

        yield connection


def convertElements(elements, clientsInOpsi, opsiClientGroupMappings,
                    session, attributeMapping, rootGroup, parentGroupId=None,
                    updateExistingClients=False,
                    preferLocationFromDrectory=False, **_):

    def isComputer(dn):
        computerDetected = False
        allElements = set(str(dn).split(","))
        for element in allElements:
            key, value = element.split('=', 1)

            if key.lower() == 'cn' and value.lower().startswith('computer'):
                computerDetected = True
                break

        LOGGER.debug("{0} is computer? {1}".format(dn, computerDetected))
        return computerDetected

    def getParent(current, groups):
        LOGGER.debug("Groups {0} ".format(str(groups)))
        if current in groups:
            LOGGER.debug('{0!r} in {1}'.format(current, groups))
            if groups.index(current) <= len(groups) - 1:
                LOGGER.debug('{0} has len {1}'.format(groups, len(groups)))
                LOGGER.debug('{0!r} has index {1}'.format(current, groups.index(current)))
                if groups.index(current) + 1 <= len(groups) - 1:
                    return groups[groups.index(current) + 1]
        return rootGroup

    def groupExistsInOpsi(groupname):
        exists = False
        for group in groupsInOpsi:
            if group["type"] == "HostGroup":
                if groupname.lower() == group["ident"]:
                    exists = True
        LOGGER.debug("groupExistsInOpsi {0!r}: {1}".format(groupname, exists))
        return exists

    def groupHasRightParent(groupId, parentGroupId):
        LOGGER.debug("Checking if {0!r} has parent {1!r}".format(groupId, parentGroupId))
        for group in groupsInOpsi:
            if group["type"] == "HostGroup":
                if groupId.lower() == group["id"]:
                    if parentGroupId.lower() == group["parentGroupId"]:
                        LOGGER.debug("Parent found ")
                        return True
        LOGGER.debug("Parent not found ")
        return False

    def createHostGroup(element, name, parentGroupId=None):
        if parentGroupId is None:
            parentGroupId = rootGroup
        if name == parentGroupId:
            raise Exception("Groupname is same as parent groupname. forbidden")
        return {
            'method': "group_insertObject",
            'params': [{
                "ident": name,
                "description": convertToStringIfNotNone(element, attributeMapping['group']["description"]),
                "notes": convertToStringIfNotNone(element, attributeMapping['group']["notes"]),
                "parentGroupId": parentGroupId,
                "type": "HostGroup"
            }],
            'id': 1}

    def createObj2Grp(groupname, clientname):
        return {
            'method': "objectToGroup_createObjects",
            'params': [{
                "groupId": groupname,
                "objectId": clientname,
                "groupType": "HostGroup",
                "type": "ObjectToGroup"
            }],
            'id': 1}

    def listReplace(llist, find, replaceWith):
        for index, item in enumerate(llist):
            if item == find:
                LOGGER.debug('replacing {0!r} with {1!r}'.format(llist[index], replaceWith))
                llist[index] = replaceWith
        return llist

    def checkRightGroup(parentlist, groupName, element, parentGroupId):
        if not groupExistsInOpsi(groupName):
            LOGGER.info("Group {0!r} does not exist in opsi. Creating.".format(groupName))
            LOGGER.debug("parentlist {0}, ".format(parentlist))
            parentGroupId = getParent(groupName, parentlist)
            return parentlist, createHostGroup(element, groupName, parentGroupId)
        else:
            LOGGER.info("Group {0!r} exists in opsi.".format(groupName))
            if parentGroupId is None:
                parentGroupId = rootGroup

            if parentGroupId and not groupHasRightParent(groupName, parentGroupId):
                while groupExistsInOpsi(groupName):
                    if groupHasRightParent(groupName, parentGroupId):
                        LOGGER.debug("Group {0!r} has right parent {1!r}".format(groupName, parentGroupId))
                        break
                    else:
                        LOGGER.debug(
                            "Group {0!r} has not right parent {1!r}."
                            "Appending {2!r} to {0!r}"
                            .format(
                                groupName,
                                parentGroupId,
                                NAME_ADDITION
                            )
                        )
                        parentlist = listReplace(parentlist, groupName, groupName+NAME_ADDITION)
                        groupName = ''.join((groupName, NAME_ADDITION))
                if not groupExistsInOpsi(groupName):
                    LOGGER.debug("Group {0!r} will be created under {1!r}".format(groupName, parentGroupId))
                    return parentlist, createHostGroup(element, groupName, parentGroupId=parentGroupId)
            else:
                LOGGER.debug(
                    "Group {groupname!r} does exist in opsi. "
                    "Parent is {parent!r}.".format(
                        groupname=groupName,
                        parent=parentGroupId
                    )
                )
                if parentGroupId:
                    LOGGER.debug(
                        "Has the group {groupname!r} the right parent? {0}".format(
                            groupHasRightParent(groupName, parentGroupId),
                            groupname=groupName
                        )
                    )
            return parentlist, None

    def deleteStrObj2Grp(objectToGroup):
        return {
           'method': "objectToGroup_deleteObjects",
           'params': [json.dumps(objectToGroup)],
           'id': 1}

    def deleteObj2Grp(clientname):
        def isFromRootDir(groupname):
            if groupname is None:
                return False
            elif groupname == rootGroup:
                return True

            for grp in groupsInOpsi:
                if groupname == grp["id"]:
                    result = isFromRootDir(grp["parentGroupId"])
                    LOGGER.debug("isFromRootDir parentGroupId {0!r}: {1}".format(grp["parentGroupId"], result))
                    return result

        for clientGroupMapping in opsiClientGroupMappings:
            if clientname == clientGroupMapping["objectId"]:
                groupname = clientGroupMapping["groupId"]
                LOGGER.debug("Found group {1!r} for client {0!r}".format(clientname, groupname))
                if isFromRootDir(groupname):
                    return deleteStrObj2Grp(clientGroupMapping)

    groupsInOpsi = getGroupsFromOpsi(session)
    opsiClients = set(clientsInOpsi)
    for element in elements:
        LOGGER.debug("Next element: {0!r}".format(element))

        try:
            dn = getDNFromElement(element)

            try:
                category = getAttributeFromElement(element, 'objectCategory', flatten=True)
                fullDN = "{0},{1}".format(category, dn)
            except KeyError as kerr:
                LOGGER.debug("Creating full DN failed: {0!r}".format(kerr))
                fullDN = dn

            parentlist = getGrouplist(fullDN)
            if isComputer(fullDN):
                try:
                    name = getAttributeFromElement(element, attributeMapping['client']['id'], flatten=True)
                except KeyError as kerr:
                    LOGGER.debug("{0!r} during reading of the attribute for the client id.".format(kerr))
                    LOGGER.critical("Unable to read the attribute {attr!r} specified for the client ID. Aborting.".format(attr=attributeMapping['client']['id']))
                    raise kerr

                if not isFqdn(name):
                    LOGGER.debug("Need to create a fqdn...")
                    fqdn = "{0}.{1}".format(name, getDomain(dn))
                else:
                    fqdn = name

                fqdn = fqdn.lower()  # opsi expected FQDN in lowercase.
                LOGGER.debug("Using {0!r} as fqdn".format(fqdn))

                if fqdn in opsiClients:
                    LOGGER.info("Client {0!r} exists in opsi".format(fqdn))
                    if updateExistingClients:
                        LOGGER.info("Updating existing client {0!r}".format(fqdn))
                        if preferLocationFromDrectory:
                            LOGGER.debug("Removing existing object2Groups {0!r}".format(fqdn))
                            yield deleteObj2Grp(fqdn)

                        LOGGER.info("Checking groups of client {0!r}".format(fqdn))
                        if parentlist:
                            for groupName in parentlist:
                                parentGroupId = getParent(groupName, parentlist)
                                LOGGER.debug("found parentId {0!r}".format(parentGroupId))
                                pparentlist, result = checkRightGroup(parentlist, groupName, element, parentGroupId)
                                groupsInOpsi = getGroupsFromOpsi(session)
                                if pparentlist is not None:
                                    parentlist = pparentlist
                                if result is not None:
                                    LOGGER.debug("Client {0!r}, {1}, {2}".format(fqdn, parentlist, result))
                                    yield result

                            yield createObj2Grp(parentlist[0], fqdn)

                        yield getClientUpdateAction(session, fqdn, element, attributeMapping['client'])
                    else:
                        LOGGER.debug("Skipped existing client {!r}".format(element))
                else:
                    LOGGER.info("Client {0!r} does not exist in opsi.".format(fqdn))
                    yield {
                        'method': 'host_createOpsiClient',
                        'params': [
                            fqdn,
                            None,  # HostKey - not setting one
                            convertToStringIfNotNone(element, attributeMapping['client']["description"]),
                            convertToStringIfNotNone(element, attributeMapping['client']["notes"]),
                            convertToStringIfNotNone(element, attributeMapping['client']["hardwareAddress"]),
                            convertToStringIfNotNone(element, attributeMapping['client']["ipAddress"]),
                            convertToStringIfNotNone(element, attributeMapping['client']["inventoryNumber"]),
                            convertToStringIfNotNone(element, attributeMapping['client']["oneTimePassword"]),
                        ],
                        'id': 1}

                    LOGGER.info("Setting groups for new client {0!r}".format(fqdn))
                    if parentlist:
                        for groupName in parentlist:
                            parentGroupId = getParent(groupName, parentlist)
                            LOGGER.debug("Found parentId {0!r}".format(parentGroupId))
                            pparentlist, result = checkRightGroup(parentlist, groupName, element, parentGroupId)
                            groupsInOpsi = getGroupsFromOpsi(session)
                            if pparentlist is not None:
                                parentlist = pparentlist
                            if result is not None:
                                LOGGER.debug("Client {0!r}, {1}, {2}".format(fqdn, parentlist, result))
                                yield result

                        yield createObj2Grp(parentlist[0], fqdn)
            else:
                LOGGER.info("Element {0} is not a computer. Syncing currently not supported.".format(element['dn']))
        except LDAPKeyError as error:
            LOGGER.warning("Processing {} failed: {}".format(element, error))
            LOGGER.debug("Full exception: {!r}".format(error))


def getDNFromElement(element):
    try:
        return element['dn']
    except KeyError as kerr:
        LOGGER.debug("Failed to get 'dn' from {0!r}: {error!r}".format(element, error=kerr))
        LOGGER.critical("Unable to get the DN for the current element!")
        raise kerr


def getGrouplist(dn):
    """
    Gets the groups from a DN.

    As a group every Organisational Unit (OU) qualifies.

    :type dn: str
    :returntype: list
    :return: groups in the
    """
    organisationalUnits = ("OU=", "ou=")
    groups = []
    for element in dn.split(","):
        for organisationalUnit in organisationalUnits:
            if organisationalUnit in element:
                groups.append(element.replace(organisationalUnit, ""))
    LOGGER.debug("Groups {0!r} extracted from {1}".format(groups, dn))
    return groups


def getAttributeFromElement(element, attribute, flatten=True):
    """
    Return the value for `attribute` in element.

    This is designed to work with different versions of dicts created
    from different versions of ldap3.

    :param element: The element to search the attribute for.
    :type element: dict
    :param attribute: The attribute to search for.
    :type attribute: str
    :param flatten: If the attributes holds many values this selects the first value to be returned.
    :type flatten: bool
    :raises KeyError: If no matching attribute is found.
    """
    try:
        attr = element['attributes'][attribute]
    except KeyError as kerr:
        LOGGER.debug("Trying direct access to {0!r} after accessing in 'attributes' failed with {1!r}".format(attribute, kerr))
        attr = element[attribute]

    if flatten and isinstance(attr, list):
        return attr[0]

    return attr


def isFqdn(name):
    return name.count('.') >= 2


def getDomain(dn):
    """
    Get the domain from the given dn.

    :type dn: str
    :return: the domain from the DN in format `domain.tld`
    :returntype: str
    """
    domainComponents = ("DC=", "dc=")
    domainParts = []
    for element in dn.split(","):
        for domainComponent in domainComponents:
            if element.startswith(domainComponent):
                domainParts.append(element.replace(domainComponent, ""))

    domain = ".".join(domainParts)
    LOGGER.debug("domain {0!r} created from {1}".format(domain, dn))
    return domain


def convertToStringIfNotNone(element, attribute):
    """
    Read the value for `attribute` from `element`.

    If the value is not `None`, then it will be explicitely converted to str.
    If there is no attribute found `None` will be returned.
    If no attribute is given `None` will be returned.

    :type attribute: str
    :param attribute: The attribute to search for.
    :returntype: str or None
    """
    if not attribute:
        return None

    try:
        attr = getAttributeFromElement(element, attribute, flatten=True)
        if attr is not None:
            return str(attr)
    except KeyError as kerr:
        LOGGER.debug("No {attr!r} in {elem!r}? Received {err!r}".format(attr=attribute, elem=element, err=kerr))

    return None


def getClientUpdateAction(session, clientId, element, clientAttributeMapping):
    opsiHost = getHostFromOpsi(session, clientId)

    updateNecessary = False
    for opsiAttribute, dirAttribute in clientAttributeMapping.items():
        if not dirAttribute or opsiAttribute == 'id':
            # We limit the updating to the attributes that have a
            # mapping and we also do not want override the id because
            # changing the ID would result in an different opsi object.
            continue

        attr = opsiAttribute

        valueInOpsi = opsiHost[attr]
        valueInDirectory = convertToStringIfNotNone(element, clientAttributeMapping[opsiAttribute])

        LOGGER.debug("Comparing {0!r} (opsi) and {1!r} (directory)".format(valueInOpsi, valueInDirectory))
        if areValuesDifferent(valueInOpsi, valueInDirectory):
            LOGGER.debug("Difference detected. Updating {attr} of our opsi client.".format(attr=opsiAttribute))
            opsiHost[opsiAttribute] = valueInDirectory
            updateNecessary = True

    if updateNecessary:
        LOGGER.info("Updating attributes of client {0!r}".format(clientId))
        return {
            'method': 'host_insertObject',  # use insert to make sure that values with None are updated aswell
            'params': [opsiHost],
            'id': 1
        }


def areValuesDifferent(firstValue, secondValue):
    """
    Returns `True` if the values differ.

    This will not compare two falsy values, i.e. '' and None.
    """
    return any((firstValue, secondValue)) and firstValue != secondValue


def getHostFromOpsi(session, clientId):
    hosts = getFromOpsi(session, 'host_getHashes', [None, {"type": "OpsiClient", "id": clientId}])
    return hosts[0]


def updateOpsi(session, actions, writeChanges=False):
    processableActions = (action for action in actions if action is not None)

    actionCounter = 0
    for actionCounter, action in enumerate(processableActions, start=1):
        LOGGER.info('Processing {!r}'.format(action['method']))
        if writeChanges:
            response = session.get(params=json.dumps(action))
            content = response.json()
            if content['error']:
                raise RuntimeError("Call to opsi failed: {class} - {message}".format(**content['error']))
            LOGGER.debug("Result from call to opsi: {0}".format(content["result"]))
            LOGGER.info('Processing {!r} finished'.format(action['method']))

    LOGGER.debug("Processed {0} action(s) in opsi.".format(actionCounter))


def getIdsOfClientsInOpsi(session):
    return set(getFromOpsi(session, "host_getIdents"))


def getClientGroupMapping(session):
    return getFromOpsi(session, "objectToGroup_getObjects")


def getGroupsFromOpsi(session):
    return getFromOpsi(session, "group_getObjects")


def getFromOpsi(session, method, params=None):
    params = json.dumps({"method": method, "id": 1, "params": params or []})
    response = session.get(params=params)
    content = response.json()
    if content['error']:
        raise RuntimeError("Call to opsi failed: {class} - {message}".format(**content['error']))
    LOGGER.debug("get_from_opsi {0}: {1}".format(method, content["result"]))
    return content['result']


if __name__ == '__main__':
    runConnector()
